import importlib
import inspect
import traceback

import clilabs


__all__ = ['debug', 'help']


def _modhelp(modname):
    filename = importlib.find_loader(modname).get_filename()
    moddoc = filedoc(filename)

    if moddoc:
        print('Module docstring:', moddoc)

    try:
        mod = importlib.import_module(modname)
    except ImportError:
        traceback.print_exc()
        print(f'Could not import module: {modname}')
    else:
        callables = clilabs.callables(mod)
        if callables:
            print(f'Callables found in: {filename}')
            print("\n".join(callables))
        else:
            print(f'No callable found in {filename}')

    if not moddoc and not callables:
        print('No help found')


def filedoc(filepath):
    co = compile(open(filepath).read(), filepath, 'exec')
    if co.co_consts and isinstance(co.co_consts[0], str):
        docstring = co.co_consts[0]
    else:
        docstring = None
    return docstring


def help(cb=None):
    """
    Get help for a callable, or list callables for a module.

    Examples:

    # Print module docstring and list of callables:
    clilabs help your.mod

    # Print callable docstring
    clilabs help your.mod:func
    """
    if not cb:
        cb = 'clilabs:cli'

    modname, funcname = clilabs.funcexpand(cb)

    try:
        cb = clilabs.modfuncimp(modname, funcname)
    except ImportError:
        if ':' not in cb:
            _modhelp(modname)
    else:
        funcdoc = inspect.getdoc(cb)
        if funcdoc:
            print(funcdoc)
        else:
            print(f'No docstring found for {modname}:{funcname}')
            _modhelp(modname)


def debug(*args, **kwargs):
    """Print debug output for a command line.

    The debug function is made to dump the result of the clilabs parser.
    It will show what callable and arguments it will use.
    You will see that the following are not the same, as stated in the
    tutorial.

    clilabs debug your:func -x 12
    clilabs debug your:func -x=12
    """
    try:
        cb = clilabs.funcimp(args[0])
    except ImportError:
        cb = args[0]
        print(f'Could not import {cb}')
    else:
        print(f'Callable: {cb}')
        print(f'Callable path: {cb.__code__.co_filename}')

    print(f'Args: {args[1:]}')
    print(f'Kwargs: {kwargs}')
    print(f'Context args: {clilabs.context.args}')
    print(f'Context kwagrs: {clilabs.context.kwargs}')



"""Builtin callables for Django framework.

It requires DJANGO_SETTINGS_MODULE env var, and does not require to be in
INSTALLED_APPS because automation software should be made automatically
available and adding to INSTALLED_APPS is a manual step.
"""
import glob
import os
import pprint
import re
import sys
import traceback

import django
from django.apps import apps

from tabulate import tabulate


found = glob.glob('**/manage.py', recursive=True)
if found and 'DJANGO_SETTINGS_MODULE' not in os.environ:
    with open(found[0], 'r') as f:
        for line in f.readlines():
            m = re.match('.*[\'"]([^\'"]*.settings[^\'"]*)[\'"]', line)
            if m:
                mod = m.group(1)
                print(f'Auto-detected DJANGO_SETTINGS_MODULE={mod}')
                print('If incorrect, set DJANGO_SETTINGS_MODULE env var')
                os.environ['DJANGO_SETTINGS_MODULE'] = mod
                break


try:
    django.setup()
except Exception:
    print('Setting up django has failed !')
    print('DJANGO_SETTINGS_MODULE env var not set !')
    if 'DJANGO_SETTINGS_MODULE' in os.environ:
        print(f'DJANGO_SETTINGS_MODULE={os.getenv("DJANGO_SETTINGS_MODULE")}')
        traceback.print_exc()


def _modeldata(obj, keys=None):
    keys = keys or [
        k for k in obj.__dict__.keys()
        if not k.startswith('_')
    ]

    return {k: getattr(obj, k) for k in keys}


def _printqs(qs, keys=None):
    keys = keys or None
    header = sorted(list(_modeldata(qs[0], keys).keys()))
    print(tabulate([header] + [
        [getattr(i, k) for k in header]
        for i in qs
    ]))


def create(modelname, *args, **kwargs):
    """Idempotent create function.

    First argument must be model name, for apps.get_model.
    With only keyword arguments, it will pass them to create().
    If you pass arguments, it will use update_or_create, passing
    any keyword argument name as defaults to update_or_create
    instead of kwarg.

    # Create a user, not idempotent
    playlabs +django:create auth.user username=foo email=joe@example.com

    # Create or update a user based on email, idempotent yay !
    playlabs +django:create auth.user email username=foo email=joe@example.com

    # oh, and with settings.* support for your model swapping fun hacks ;)
    playlabs +django:create settings.AUTH_USER_MODEL ...
    """
    if modelname.startswith('settings.'):
        from django.conf import settings
        modelname = getattr(settings, modelname.split('.')[1])
    model = apps.get_model(modelname)

    if not args:
        obj = model.objects.create(**kwargs)
        created = True
    else:
        defaults = {}
        for key, value in kwargs.copy().items():
            if key not in args:
                defaults[key] = kwargs.pop(key)
        obj, created = model.objects.update_or_create(defaults, **kwargs)

    print(tabulate([
        (k, v)
        for k, v in _modeldata(obj).items()
    ]))


def ls(modelname, *args, **kwargs):
    """Search models

    kwargs are passed to filter.
    It shows all fields by default, you can restrict them with args.

    Show username and email for superusers:

    clilabs +django:ls auth.user is_superuser=1 username email
    """

    model = apps.get_model(modelname)
    models = model.objects.filter(**kwargs)
    if not models:
        print('No result found !')
        sys.exit(0)

    _printqs(models, args)


def delete(modelname, **kwargs):
    model = apps.get_model(modelname)
    qs = model.objects.filter(**kwargs)
    if not qs:
        print('No model to delete !')
        return
    _printqs(qs)
    count = len(qs)
    qs.delete()
    print(f'Deleted {count} objects')


def detail(modelname, *args, **kwargs):
    """Print detail for a model.

    kwargs are passed to filter()

    clilabs +django:detail pk=123
    """
    model = apps.get_model(modelname)
    obj = model.objects.get(**kwargs)
    print(tabulate([
        (k, v)
        for k, v in _modeldata(obj).items()
        if k in args or not args
    ]))


def chpasswd(password, **kwargs):
    """Change the password for user.

    It takes the password as argument, that you can use `-` for stdin.
    All kwargs will be passed to get()

    Example:

        clilabs +django:chpasswd username=... thepassword
        echo thepassword | clilabs +django:chpasswd username=... -
    """

    from django.conf import settings
    model = apps.get_model(settings.AUTH_USER_MODEL)
    user = model.objects.get(**kwargs)
    user.set_password(password)
    user.save()
    print('Password updated !')


def settings(*names):
    """Show settings from django.

    How many times have you done the following ?

        python manage.py shell
        from django.conf import settings
        settings.DATABASES # or something

    Well it's over now ! Try this instead:

        clilabs +django:settings DATABASES INSTALLED_APPS # etc
    """
    from django.conf import settings

    for name in names:
        print(f'{name}={pprint.pformat(getattr(settings, name))}')


def main(*args, **kwargs):
    print('For help, run clilabs help +django')



import io
import clilabs
import clilabs.builtins


class Foo:
    bar = {'baz': [lambda *a, **k: 'test']}


CB_NAME = 'clilabs.test_clilabs:Foo.bar.baz.0'


def test_funcimp():
    result = clilabs.funcimp(CB_NAME)
    assert result == Foo.bar['baz'][0]


def test_expand(monkeypatch):
    monkeypatch.setattr('sys.stdin', io.StringIO('in'))
    result = clilabs.expand('-', '-a', '--b=1', 'c', 'd=1')
    assert result == (['in', 'c'], {'d': '1'})


def test_cli():
    result = clilabs.cli('clilabs', CB_NAME, '1')
    assert result == 'test'


def test_help(capsys):
    clilabs.builtins.help()
    assert 'clilabs' in capsys.readouterr().out


def test_help_builtin(capsys):
    clilabs.builtins.help('help')
    assert 'help' in capsys.readouterr().out


def test_help_module(capsys):
    clilabs.builtins.help('clilabs')
    assert 'cli' in capsys.readouterr().out



import inspect
import importlib
import sys


def cli(*argv):
    '''clilabs automates python callables parametered calls.

    Things starting with - will arrive in clilabs.context.

    Examples:

        clilabs help your.mod:funcname to get its docstring.
        clilabs debug your.mod -a --b --something='to see' how it=parses
        clilabs your.mod:funcname with your=args
        clilabs help clilabs.django
    '''
    argv = argv if argv else sys.argv
    if len(argv) < 2:
        argv.append('help')

    cb = funcimp(argv[1])
    args, kwargs = expand(*argv[2:])
    return cb(*args, **kwargs)


def callables(mod):
    return [
        i[0]
        for i in inspect.getmembers(mod)
        if callable(getattr(mod, i[0]))
        and not i[0].startswith('_')
    ]


def funcexpand(callback):
    import clilabs.builtins
    builtins = callables(clilabs.builtins)

    if callback.startswith('+'):
        callback = f'clilabs.{callback[1:]}'

    if callback in builtins:
        funcname = callback
        modname = 'clilabs.builtins'
    elif ':' not in callback:
        funcname = 'main'
        modname = callback
    else:
        modname, funcname = callback.split(':')
        if not modname:
            modname = 'clilabs.builtins'

    return modname, funcname


def modfuncimp(modname, funcname):
    ret = importlib.import_module(modname)
    for part in funcname.split('.'):
        if isinstance(ret, dict) and part in ret:
            ret = ret.get(part)
        elif isinstance(ret, list) and part.isnumeric():
            ret = ret[int(part)]
        else:
            ret = getattr(ret, part, None)

        if ret is None:
            raise ImportError(f'{part} is None')

    return ret


def funcimp(callback):
    return modfuncimp(*funcexpand(callback))


def expand(*argvs):
    args, kwargs = list(), dict()

    for argv in argvs:
        if argv == '-':
            args.append(sys.stdin.read().strip())
            continue

        if argv.startswith('-'):
            continue

        if '=' in argv:
            name, value = argv.split('=')
            if value == '-':
                value = sys.stdin.read().strip()
            kwargs[name] = value
        else:
            args.append(argv)

    return args, kwargs


class Context:
    def __init__(self, *args, **kwargs):
        self.args = list(args)
        self.kwargs = kwargs

    @classmethod
    def factory(cls, argvs):
        context = cls()

        for argv in argvs:
            if not argv.startswith('-'):
                continue

            if argv == '--':
                context.args.append(sys.stdin.read().strip())
                continue

            argv = argv.lstrip('-')

            if '=' in argv:
                key, value = argv.split('=')
                if value == '-':
                    value = sys.stdin.read().strip()
                context.kwargs[key] = value

            else:
                context.args.append(argv)

        return context


context = Context.factory(sys.argv)



"""
Automate sentry-cli cooking for uploading sourcemaps.
"""
import glob
import os
import subprocess
import sys


def _run(cmd):
    print('+', ' '.join(cmd))

    proc = subprocess.Popen(
        cmd,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )

    try:
        outs, errs = proc.communicate(timeout=30)
    except subprocess.TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()

    if outs:
        print(outs.decode('utf8'))
    if errs:
        print(errs.decode('utf8'))

    return proc.returncode


def _getenv(name):
    value = os.getenv(name)
    if not value:
        print(f'{name} not set, aborting')
        sys.exit(1)
    return value


def release_delete():
    """
    Equivalent to:

        sentry-cli releases delete $GIT_COMMIT

    Except your exit code will always be 0 (idempotent-ish).
    """
    version = _getenv('GIT_COMMIT')
    cmd = ['sentry-cli', 'releases', 'delete', version]
    _run(cmd)


def release_new():
    """
    Equivalent to:

        sentry-cli releases new -p $SENTRY_PROJECT $GIT_COMMIT
    """
    projects = _getenv('SENTRY_PROJECT')
    version = _getenv('GIT_COMMIT')

    cmd = ['sentry-cli', 'releases', 'new', version]
    for p in projects.split(' '):
        cmd += ['-p', p]

    return _run(cmd)


def release_set_commits():
    """
    Equivalent to:

        sentry-cli releases set-commits $GIT_COMMIT --commit $REPO@$GIT_COMMIT
    """
    repo = _getenv('REPO')
    version = _getenv('GIT_COMMIT')

    cmd = ['sentry-cli', 'releases', 'set-commits', version]
    cmd += ['--commit', f'{repo}@{version}']
    _run(cmd)


def release_upload_js(*dirs):
    """
    Equivalent to something like:

    for i in $(find $STATIC_ROOT -type f -name '*.js' ! -name '*.test.js'); do
      sentry-cli releases files $V upload $i $STATIC_URL/${i##$STATIC_ROOT}
    done

    However, it will recursively add directories passed as arguments.

    When it builds the url of the file, it will strip STATIC_ROOT, and prepend
    STATIC_URL: it doesn't have to be the STATIC_ROOT of your django project,
    you can run this in each app you want to upload sourcemaps for, since this
    step is only needed to get sourcemaps if, for example, your environment
    is being HTTP Basic auth and sentry can't download them by itself.
    """
    version = _getenv('GIT_COMMIT')
    static_url = _getenv('STATIC_URL')
    base_cmd = ['sentry-cli', 'releases', 'files', version, 'upload']

    curdir = os.getcwd()
    for d in dirs:
        os.chdir(d)
        for i in glob.iglob('**/*.js', recursive=True):
            _run(base_cmd + [i, os.path.join(static_url, i)])
    os.chdir(curdir)


def release_upload_sourcemaps(*dirs):
    """Equivalent to:

    for i in $(find $STATIC_ROOT -type f -name '*.js.map'); do
      sentry-cli releases files $GIT_COMMIT upload-sourcemaps --validate $i
    done
    """
    version = _getenv('GIT_COMMIT')
    static_url = _getenv('STATIC_URL')
    base_cmd = [
        'sentry-cli',
        'releases',
        'files',
        version,
        'upload-sourcemaps',
        '--validate'
    ]

    curdir = os.getcwd()
    for d in dirs:
        os.chdir(d)
        for i in glob.iglob('**/*.js.map', recursive=True):
            prefix = os.path.join(static_url, '/'.join(i.split('/')[:-1]))
            _run(base_cmd + ['--url-prefix', prefix, i])
    os.chdir(curdir)


def release_finalize():
    """Equivalent to:

    sentry-cli releases finalize $GIT_COMMIT
    """
    version = _getenv('GIT_COMMIT')
    cmd = ['sentry-cli', 'releases', 'finalize', version]
    return _run(cmd)


def release(*dirs):
    """Automates all the above ! VICTORY !"""
    if os.getenv('SENTRY_OVERWRITE', False):
        release_delete()
    release_new()
    release_set_commits()
    release_upload_js(*dirs)
    release_upload_sourcemaps(*dirs)
    release_finalize()



from setuptools import setup
import os


# Utility function to read the README file.
# Used for the long_description. It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


VERSION = '1.1.1'


setup(
    name='clilabs',
    version=VERSION if '@' not in VERSION else 'dev',
    description='Cheap CLI framework, gives rich commands for Django',
    author='James Pic',
    author_email='jamespic@gmail.com',
    url='https://yourlabs.io/oss/clilabs',
    packages=['clilabs'],
    include_package_data=True,
    long_description=read('README.rst'),
    license='MIT',
    keywords='django cli',
    entry_points={
        'console_scripts': [
            'clilabs = clilabs:cli',
        ],
    },
    install_requires=['tabulate'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Software Development',
        'Topic :: System',
        'Topic :: Terminals',
    ],
    python_requires='>=3',
)
